from hhat_new_ast import (
    Main as NewMain,
    Expr as NewExpr,
    ManyExpr as NewManyExpr,
    Operation as NewOperation,
    Array as NewArray,
    Id as NewId,
    Literal as NewLiteral
)
from arpeggio import visit_parse_tree, PTNodeVisitor
from arpeggio.cleanpeg import ParserPEG


class CST(PTNodeVisitor):
    def __init__(self, defaults=True, **kwargs):
        super().__init__(defaults=defaults, **kwargs)

    def visit_program(self, n, k):
        # return Program(*k)
        # return NewMain(*k)
        res = tuple(NewManyExpr(p) for p in k)
        return NewMain(*res)

    def visit_exprs(self, n, k):
        # return Expr(*k)
        return NewExpr(*k)

    def visit_expr(self, n, k):
        if len(k) > 1:
            # return Expr(*k)
            return NewManyExpr(*k)  # NewExpr(*k)
        return k[0]

    def visit_operation(self, n, k):
        if len(k) > 1:
            # return Operation(k[0], *k[1:])
            return NewOperation(k[0], *k[1:])
        # return Operation(k[0])
        return NewOperation(k[0])

    def visit_array(self, n, k):
        # return Array(*k)
        return NewArray(*k)

    def visit_id(self, n, k):
        # return Id(n.value)
        return NewId(n.value)

    def visit_literal(self, n, k):
        return k[0]

    def visit_BOOL(self, n, k):
        # return Literal(n.value, "bool")
        return NewLiteral(n.value, "bool")

    def visit_INT(self, n, k):
        # return Literal(n.value, "int")
        return NewLiteral(n.value, "int")


def parse_code(code):
    peg_grammar = open("grammar.peg", "r").read()
    parsed_code = ParserPEG(peg_grammar, "program", reduce_tree=False)
    pt = parsed_code.parse(code)
    return visit_parse_tree(pt, CST())



from abc import ABC
from typing import Any
from hhat_literal_define import (literal_bool_define, literal_int_define)


literal_dict = dict(
    bool=literal_bool_define,
    int=literal_int_define
)


class AOT(ABC):
    def __init__(self, token: str, aot_type: str = ""):
        self.token = token
        self.type = aot_type

    def __repr__(self) -> str:
        return self.token


class AST(ABC):
    def __init__(
            self,
            *args: Any,
            node: AOT | None = None,
            ast_type: str = "",
            is_concurrent: bool = False,
    ):
        self.node = node or ""
        self.type = ast_type
        self.edges = args
        self.is_concurrent = is_concurrent

    def __len__(self):
        return len(self.edges)

    def __iter__(self):
        yield from self.edges

    def __repr__(self) -> str:
        token = self.node.token if self.node else ""
        args = " ".join(str(k) for k in self.edges)
        return token + (f"({args})" if args else "")


class Literal(AOT):
    def __init__(self, token: str, lit_type: str = ""):
        super().__init__(token, lit_type)
        self.value = literal_dict[self.type](self.token)


class Id(AOT):
    def __init__(self, token: str, aot_type: str = "id"):
        super().__init__(token, aot_type)
        self.value = self.token


class Array(AST):
    def __init__(self, *values: Any):
        super().__init__(ast_type="array", *values)


class Expr(AST):
    def __init__(self, *values: Any):
        super().__init__(ast_type="expr", is_concurrent=True, *values)


class ManyExpr(AST):
    def __init__(self, *values: Any):
        super().__init__(ast_type="many-expr", is_concurrent=False, *values)


class Operation(AST):
    def __init__(self, oper_token: AOT | str, *args: Any):
        if isinstance(oper_token, Id):
            oper_token.type = "oper"
        else:
            oper_token = Id(token=oper_token, aot_type="oper")
        super().__init__(node=oper_token, ast_type="oper", *args)


class Main(AST):
    def __init__(self, *exprs: Any):
        super().__init__(ast_type="main", is_concurrent=True, *exprs)


class Program(AST):
    def __init__(self, *super_exprs: Any):
        super().__init__(ast_type="program", *super_exprs)



from hhat_parsing import parse_code
from hhat_pre_eval import PreEval
from hhat_eval import Eval
from hhat_semantics import Analysis


def parsing(c):
    pc_ = parse_code(c)
    print(f"- code:\n{c}\n")
    print(f"- parsed code:\n{pc_}\n")
    return pc_


def pre_evaluating(pc):
    # pev_ = PreEval(pc)
    # res_ = pev_.run()
    print(f"- pre-evaluation:")
    analysis = Analysis(pc)
    res_ = analysis.run()
    print("\n")
    return res_


def evaluating(c):
    ev_ = Eval(c)
    print("- executing code:\n")
    ev_.run()


def run_codes(c):
    pc_ = parsing(c)
    pev_ = pre_evaluating(pc_)
    print("-" * 80)
    evaluating(pev_)


code_list = [
    """
    [2 3 4]:(sum times):print
    [10 20]:sum(2):print
    [50 60]:(sum:n times(n):m times):print
    """,
    # "[0 1]:sum [2 3 4]:times [5 6]:sum:print [7 8]:(sum times):print",
    # "[0 1]:(@init:@q1 @sync(@q1):@q2)",
    # "[0 1 2 3]:(@init:@ampl:@q1 @sync(@q1):@fflip(pi over(pi 2)):@q2)"
]


if __name__ == "__main__":
    print("***[START]***")
    print("="*80)
    for code in code_list:
        # parsing(code)
        run_codes(code)
        print('='*80)
    print("***[END]***")



from hhat_new_ast import (
    AST,
    Main,
    Expr,
    ManyExpr,
    Literal,
    Array,
    Operation
)
from hhat_memory import R


class Analysis:
    def __init__(self, parsed_code: AST):
        self.code = parsed_code

    def run(self):
        res = analyze(self.code)
        print(res)
        return res


def iter_analyze(code_: AST, role: str = ""):
    return tuple(analyze(code_=k, role=role) for k in code_)


def analyze(code_: AST, role: str = ""):
    match code_:
        case Expr():
            # print("* expr found; new concurrent task!")
            # print("  - expr start")
            res = iter_analyze(code_, role)
            # print("  - expr end")
            return R(
                ast_type=code_.type,
                value=res,
                is_concurrent=False,
                role="",
                execute_after=None,
            )
        case ManyExpr():
            # print(f"* entered many exprs ==> {code_}")
            # print("  - many-exprs enter")
            res = iter_analyze(code_, role)
            # print("  - many-exprs end")
            return R(
                ast_type=code_.type,
                value=res,
                is_concurrent=True,
                role="",
                execute_after=None,
            )
        case Literal():
            # print(f"* literal ==> {code_}")
            return code_
        case Array():
            # print(f"* array -> {code_}")
            # print("  - array start")
            res = iter_analyze(code_, role)
            # print("  - array end")
            return R(
                ast_type=code_.type,
                value=res,
                is_concurrent=False,
                role="",
                execute_after=None
            )
        case Operation():
            # print(f"* operation -> {code_}")
            res = iter_analyze(code_, role="callee")
            if res:
                return R(
                    ast_type="call",
                    value=(
                        R(
                            ast_type="oper",
                            value=code_.node,
                            is_concurrent=False,
                            role="caller",
                            execute_after=None
                        ),
                        R(
                            ast_type="args",
                            value=res,
                            is_concurrent=False,
                            role="callee",
                            execute_after=None
                        )
                    ),
                    is_concurrent=False,
                    role="",
                    execute_after=None,
                )
            return R(
                ast_type="call",
                value=(
                    R(
                        ast_type="oper",
                        value=code_.node,
                        is_concurrent=False,
                        role="caller" if not role else role,
                        execute_after=None
                    ),),
                is_concurrent=False,
                role="",
                execute_after=None,
            )
        case Main():
            res = iter_analyze(code_, role)
            return R(
                ast_type=code_.type,
                value=res,
                is_concurrent=True,
                role="",
                execute_after=None
            )
        case _:
            print(f"is {type(code_)}!")



from abc import ABC


class AST(ABC):
    def __init__(self, value, value_type, ast_type, *args):
        self.value = value
        self.type = value_type
        self.ast_type = ast_type
        self.args = args

    def __bool__(self):
        return self.value is not None

    def __len__(self):
        return len(self.args)

    def __iter__(self):
        yield from self.args

    def __repr__(self):
        value = str(self.value) if self.value else ""
        args = " ".join(str(k) for k in self.args)
        return value + (f"({args})" if args else "")


class Literal(AST):
    def __init__(self, value, literal_type):
        super().__init__(value, literal_type, "literal")


class Array(AST):
    def __init__(self, *values):
        super().__init__(None, None, "array", *values)


class Id(AST):
    def __init__(self, value):
        super().__init__(value, None, "id")


class Expr(AST):
    def __init__(self, *values):
        super().__init__("expr", None, "expr", *values)


class Operation(AST):
    def __init__(self, name, *values):
        super().__init__(name, None, "oper", *values)


class Main(AST):
    def __init__(self, *exprs):
        super().__init__("main", None, "main", *exprs)


class Program(AST):
    def __init__(self, *exprs):
        super().__init__("program", None, "program", *exprs)



from typing import Any

from hhat_lang.hhat_base_datatype import DataType, DataTypeArray


################
# SINGLE TYPES #
################

class DefaultType(DataType):
    @property
    def token(self):
        return "default-type"

    @property
    def type(self):
        return "default-type"

    def cast(self) -> Any:
        return self.value

    def __add__(self, other: Any) -> Any:
        raise ValueError(f"cannot add with {self.__class__.__name__}.")

    def __radd__(self, other: Any) -> Any:
        raise ValueError(f"cannot add with {self.__class__.__name__}.")

    def __mul__(self, other: Any) -> Any:
        raise ValueError(f"cannot multiply with {self.__class__.__name__}.")

    def __rmul__(self, other: Any) -> Any:
        raise ValueError(f"cannot multiply with {self.__class__.__name__}.")

    def __repr__(self) -> str:
        return f"<default>{self.data}"


class Bool(DataType):
    undo_bool_dict = {True: "T", False: "F"}
    convert2bool_dict = dict(T=True, F=False)

    @property
    def token(self):
        return "bool"

    @property
    def type(self):
        return "bool"

    def cast(self) -> Any:
        if self.value in self.convert2bool_dict.keys():
            return self.value
        raise ValueError(f"Wrong value for boolean: {self.value}.")

    def __add__(self, other: Any) -> Any:
        if isinstance(other, Bool):
            return Bool(self.convert2bool_dict[self.data] and self.convert2bool_dict[other.data])
        raise ValueError(f"cannot add {self.__class__.__name__} with {other.__class__.__name__}")

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, Bool):
            return Bool(other.data and self.data)
        raise ValueError(f"cannot add {self.__class__.__name__} with {other.__class__.__name__}")

    def __mul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        ...


class Int(DataType):
    @property
    def token(self):
        return "int"

    @property
    def type(self):
        return "int"

    def cast(self) -> Any:
        return int(self.value) if isinstance(self.value, str) else self.value

    def __add__(self, other: Any) -> Any:
        if isinstance(other, Int):
            return Int(self.data + other.data)
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x: self.data + x, other.data)))
        if isinstance(other, int):
            return Int(self.data + other)
        if isinstance(other, tuple):
            return IntArray(*tuple(map(lambda x: self.data + x.data, other)))
        raise ValueError(f"cannot add {self.__class__.__name__} with {other.__class__.__name__}")

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, Int):
            return Int(other.data + self.data)
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x: x + self.data, other.data)))
        if isinstance(other, int):
            return Int(other + self.data)
        if isinstance(other, tuple):
            return IntArray(*tuple(map(lambda x: x.data + self.data, other)))
        raise ValueError(f"cannot add {self.__class__.__name__} with {other.__class__.__name__}")

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, Int):
            return Int(self.data * other.data)
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x: self.data * x, other.data)))
        if isinstance(other, int):
            return Int(self.data * other)
        raise ValueError(f"cannot multiply {self.__class__.__name__} with {other.__class__.__name__}")

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, Int):
            return Int(other.data * self.data)
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x: x * self.data, other.data)))
        if isinstance(other, int):
            return Int(other * self.data)
        raise ValueError(f"cannot multiply {self.__class__.__name__} with {other.__class__.__name__}")


###############
# ARRAY TYPES #
###############

class BoolArray(DataTypeArray):
    bool_dict = dict(T=True, F=False)

    @property
    def token(self):
        return "bool-array"

    @property
    def type(self):
        return "bool"

    def cast(self) -> Any:
        return tuple(Bool(k) for k in self.value)

    def __add__(self, other: Any) -> Any:
        if isinstance(other, BoolArray):
            return BoolArray(*tuple(map(lambda x, y: x and y, self.data, other.data)))

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, BoolArray):
            return BoolArray(*tuple(map(lambda x, y: x and y, other.data, self.data)))

    def __mul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        ...


class IntArray(DataTypeArray):
    @property
    def token(self):
        return "int-array"

    @property
    def type(self):
        return "int"

    def cast(self):
        res = ()
        for k in self.value:
            if isinstance(k, IntArray):
                res += tuple(Int(p) for p in k)
            elif isinstance(k, Int):
                res += Int(k),
        return res

    def __add__(self, other: Any) -> Any:
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x, y: x + y, self.data, other.data)))
        if isinstance(other, Int):
            return IntArray(*tuple(map(lambda x: x + other.data, self.data)))
        print(f"* [add] what is other? {type(other)} {other}")

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, IntArray):
            return IntArray(*tuple(map(lambda x, y: x + y, other.data, self.data)))
        if isinstance(other, Int):
            return IntArray(*tuple(map(lambda x: other.data + x, self.data)))
        print(f"* [radd] what is other? {type(other)} {other}")

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, IntArray):
            print(f"mult int array: {self.data} ({type(self.data)}) | {other.data} ({type(other.data)})")
            return IntArray(*tuple(map(lambda x, y: x * y, self.data, other.data)))
        if isinstance(other, Int):
            return IntArray(*tuple(map(lambda x: x * other.data, self.data)))
        print(f"* [mul] mult int array: {self.data} ({type(self.data)}) | {other.data} ({type(other.data)})")

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, IntArray):
            print(f"mult int array: {self.data} ({type(self.data)}) | {other.data} ({type(other.data)})")
            return IntArray(*tuple(map(lambda x, y: x + y, other.data, self.data)))
        if isinstance(other, Int):
            return IntArray(*tuple(map(lambda x: other.data * x, self.data)))
        print(f"* [rmul] mult int array: {self.data} ({type(self.data)}) | {other.data} ({type(other.data)})")


class QArray(DataTypeArray):
    @property
    def token(self) -> str:
        return "@array"

    @property
    def type(self) -> str:
        return "@array"

    def cast(self) -> tuple[Any]:
        return tuple(k for k in self.value)

    def __add__(self, other: Any) -> Any:
        if isinstance(other, QArray):
            self.data += other,
            return self
        if isinstance(other, Int):
            # TODO: implement the casting
            return
        if isinstance(other, IntArray):
            # TODO: implement the casting
            return
        if isinstance(other, Bool):
            # TODO: implement the casting
            return
        if isinstance(other, BoolArray):
            # TODO: implement the casting
            return

        from hhat_memory import R

        if isinstance(other, R):
            self.data += other,
            return self

    def __radd__(self, other):
        if isinstance(other, QArray):
            other.data += self.data
            return other
        if isinstance(other, Int):
            # TODO: implement the casting
            return
        if isinstance(other, IntArray):
            # TODO: implement the casting
            return
        if isinstance(other, Bool):
            # TODO: implement the casting
            return
        if isinstance(other, BoolArray):
            # TODO: implement the casting
            return

        from hhat_memory import R

        if isinstance(other, R):
            self.data += other,
            return self

    def __mul__(self, other):
        if isinstance(other, QArray):
            pass
        if isinstance(other, Int):
            # TODO: implement the casting
            return
        if isinstance(other, IntArray):
            # TODO: implement the casting
            return
        if isinstance(other, Bool):
            # TODO: implement the casting
            return
        if isinstance(other, BoolArray):
            # TODO: implement the casting
            return

        from hhat_memory import R

        if isinstance(other, R):
            pass

    def __rmul__(self, other):
        if isinstance(other, QArray):
            pass
        if isinstance(other, Int):
            # TODO: implement the casting
            return
        if isinstance(other, IntArray):
            # TODO: implement the casting
            return
        if isinstance(other, Bool):
            # TODO: implement the casting
            return
        if isinstance(other, BoolArray):
            # TODO: implement the casting
            return

        from hhat_memory import R

        if isinstance(other, R):
            pass


builtin_data_types_dict = {
    "bool": Bool,
    "int": Int,
}
builtin_array_types_dict = {
    "bool": BoolArray,
    "int": IntArray,
    "@array": QArray,
}
data_types_list = tuple(builtin_data_types_dict.keys())
array_types_list = tuple(builtin_array_types_dict.keys())
